#WRITTEN BY DENE BOWDALO

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

#classification_functions.py

#module which stores functions which handle classification of stations based on metadata.

###--------------------------------------------------------------------------------------------------###
###IMPORT MODULES
###--------------------------------------------------------------------------------------------------###

import bisect
import copy
from crossnetwork_processing_functions import fractional_minutes_from_reference_to_datetime
import datetime
from GHOST_standards import standard_classification_flag_name_to_classification_code
import numpy as np
import pandas as pd

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###

def do_classifications(metadata,data):

    '''define pipeline of functions that apply classification checks to data from each station'''

    #-----------------------------------------------------#
    #HIGH ALTITUDE/STEEP TERRAIN CLASSIFICATIONS
    #-----------------------------------------------------#

    #set high altitude limit 
    high_altitude_limit = 1500 #1500 metres relative to mean sea level
    
    #High Altitude - Metadata Altitude 
    data = high_altitude_classification_by_metadata_altitude(high_altitude_limit,metadata,data)
    
    #High Altitude - ETOPO1 
    data = high_altitude_classification_by_ETOPO1(high_altitude_limit,metadata,data)
    
    #High Altitude - Iwahashi Global Landform Classification 
    #data = high_altitude_classification_by_Iwahashi(metadata,data)
    
    #High Altitude - Meybeck Global Landform Classification
    #data = high_altitude_classification_by_Meybeck(metadata,data)

    #-----------------------------------------------------#
    #NEAR COAST CLASSIFICATIONS
    #-----------------------------------------------------#
    
    #Near Coast - GSFC
    data = near_coast_classification_by_GSFC(metadata,data)
 
    #-----------------------------------------------------#
    #RURAL/URBAN CLASSIFICATIONS
    #-----------------------------------------------------#
    
    #rural/urban classifications using standardised network provided metadata
    data = rural_urban_classification_by_standardised_metadata(metadata,data)
    
    #rural/urban classifications using ANTHROME (native resolution) classification
    data = rural_urban_classification_by_anthrome(metadata,data,'native')
    
    #rural/urban classifications using ANTHROME (mode in 5km perimeter) classification
    data = rural_urban_classification_by_anthrome(metadata,data,'5km')

    #rural/urban classifications using ANTHROME (mode in 25km perimeter) classification
    data = rural_urban_classification_by_anthrome(metadata,data,'25km')

    #rural/urban classifications using TOAR classification
    #data = rural_urban_classification_by_TOAR(metadata,data)
    
    #rural/urban classifications using Joly-Peuch classification
    #data = rural_classification_by_Joly_Peuch(metadata,data)

    #-----------------------------------------------------#
    #STANDARDISED METADATA CLASSIFICATIONS
    #-----------------------------------------------------#

    #standardised network provided area classification 
    data = get_standardised_metadata_classfication('standardised_network_provided_area_classification','Area Classification Metadata',metadata,data)

    #standardised network provided station classification 
    data = get_standardised_metadata_classfication('standardised_network_provided_station_classification','Station Classification Metadata',metadata,data)
    
    #standardised network provided spatial representativity
    data = spatial_representativity_classfication(metadata,data)

    #standardised network provided terrain
    data = get_standardised_metadata_classfication('standardised_network_provided_terrain','Terrain Metadata',metadata,data)

    #standardised network provided land use
    data = get_standardised_metadata_classfication('standardised_network_provided_land_use','Land Use Metadata',metadata,data)
    
    #MODIS MCD12C1 land use
    data = get_standardised_metadata_classfication('MODIS_MCD12C1_v6_IGBP_land_use','MODIS MCD12C1 Land Use',metadata,data)

    #-----------------------------------------------------#
    #TEMPORAL CLASSIFICATIONS
    #-----------------------------------------------------#
    
    #get daytime/nighttime classification
    data = check_nighttime(metadata, data)

    #get season classification
    data = check_season(metadata, data)
    
    return data

###--------------------------------------------------------------------------------------------------###
###--------------------------------------------------------------------------------------------------###
#CLASSIFICATION FUNCTIONS

def high_altitude_classification_by_metadata_altitude(high_altitude_limit,metadata,data):

    '''Check for high altitude stations using network provided measurement altitude information
       check is done by seeing if altitude is >= high altitude limit.
    '''   
    
    #load metadata
    measurement_alts = metadata['measurement_altitude']
    update_metadata_stamp = metadata['meta_update_stamp']
    
    #define list for keeping indices associated with high altitude stations
    high_altitude_inds = np.empty(0, dtype=np.uint32)
    
    #iterate though all measurement altitude metadata
    for ii in range(len(measurement_alts)):
    
        #above high altitude limit?
        if measurement_alts[ii] >= high_altitude_limit:
            #is it last index? 
            if ii == (len(measurement_alts)-1):
                high_altitude_inds = np.append(high_altitude_inds,np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), len(data['time'])))
            else:
                high_altitude_inds = np.append(high_altitude_inds,np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), bisect.bisect_left(data['time'], update_metadata_stamp[ii+1])))
                
    #update relevant high altitude indices with classification codes 
    if len(high_altitude_inds) > 0:
        #get relevant classification flag code
        classification_flag_code = standard_classification_flag_name_to_classification_code['High Altitude - Metadata Altitude']
        for inv in high_altitude_inds:
            data['classification'][inv].append(classification_flag_code)
        
    return data 

###------------------------------------------------------------------------------------###

def high_altitude_classification_by_ETOPO1(high_altitude_limit,metadata,data):

    '''Check for high altitude stations using ETOPO1 digital elevation model gridded data.
       check is done by adding ETOPO1 altitude + sampling height to produce a measurement height, and this is checked to see if it is >= high altitude limit.
       if no sampling height is available, only ETOPO1 altitude is used. 
    '''

    #load metadata
    ETOPO1_altitudes = metadata['ETOPO1_altitude']
    sampling_heights = metadata['sampling_height']
    update_metadata_stamp = metadata['meta_update_stamp']
    
    #create measurement altitude field (adding ETOPO1 altitude + sampling height (where available))
    measurement_alts = copy.deepcopy(ETOPO1_altitudes)
    for ii in range(len(measurement_alts)):
        #if have valid sampling height and altitude, create a measurement altitude
        if (pd.isnull(sampling_heights[ii]) == False) & (pd.isnull(ETOPO1_altitudes[ii]) == False):
            measurement_alts[ii] = sampling_heights[ii] + ETOPO1_altitudes[ii]
    
    #define list for keeping indices associated with high altitude stations
    high_altitude_inds = np.empty(0, dtype=np.uint32)
    
    #iterate though all measurement altitudes
    for ii in range(len(measurement_alts)):
    
        #above high altitude limit?
        if measurement_alts[ii] >= high_altitude_limit:
            #is it last index? 
            if ii == (len(measurement_alts)-1):
                high_altitude_inds = np.append(high_altitude_inds,np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), len(data['time'])))
            else:
                high_altitude_inds = np.append(high_altitude_inds,np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), bisect.bisect_left(data['time'], update_metadata_stamp[ii+1])))
                
    #update relevant high altitude indices with classification code 
    if len(high_altitude_inds) > 0:
        #get relevant classification flag code
        classification_flag_code = standard_classification_flag_name_to_classification_code['High Altitude - ETOPO1']
        for inv in high_altitude_inds:
            data['classification'][inv].append(classification_flag_code)
        
    return data 

###------------------------------------------------------------------------------------###

def high_altitude_classification_by_Iwahashi(metadata,data):

    '''Check for high altitude stations using the European Soil Data Centre Iwahashi Global Landform Classification gridded product'''

    #load metadata
    Iwahashi_classifications = metadata['ESDAC_Iwahashi_landform_classification']
    update_metadata_stamp = metadata['meta_update_stamp']

    #define list for keeping indices associated with high altitude stations
    high_altitude_inds = np.empty(0, dtype=np.uint32)
    
    #iterate though all classifications
    #for ii in range(len(Iwahashi_classifications)):
        
        #if Iwahashi_classifications[ii] == '':
            #is it last index? 
        #    if ii == (len(Iwahashi_classifications)-1):
        #        high_altitude_inds = np.append(high_altitude_inds,np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), len(data['time'])))
        #    else:
        #        high_altitude_inds = np.append(high_altitude_inds,np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), bisect.bisect_left(data['time'], update_metadata_stamp[ii+1])))
                
    #update relevant high altitude indices with classification code  
    if len(high_altitude_inds) > 0:
        #get relevant classification flag code
        classification_flag_code = standard_classification_flag_name_to_classification_code['High Altitude - Iwahashi Global Landform Classification']
        for inv in high_altitude_inds:
            data['classification'][inv].append(classification_flag_code)
    

    return data

###------------------------------------------------------------------------------------###

def high_altitude_classification_by_Meybeck(metadata,data):

    '''Check for high altitude stations using the European Soil Data Centre Meybeck Global Landform Classification gridded product'''

    #load metadata
    Meybeck_classifications = metadata['ESDAC_Meybeck_landform_classification']
    update_metadata_stamp = metadata['meta_update_stamp']

    #define list for keeping indices associated with high altitude stations
    high_altitude_inds = np.empty(0, dtype=np.uint32)
    
    #iterate though all classifications
    #for ii in range(len(Meybeck_classifications)):
        
        #if Meybeck_classifications[ii] == '':
            #is it last index? 
        #    if ii == (len(Meybeck_classifications)-1):
        #        high_altitude_inds = np.append(high_altitude_inds,np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), len(data['time'])))
        #    else:
        #        high_altitude_inds = np.append(high_altitude_inds,np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), bisect.bisect_left(data['time'], update_metadata_stamp[ii+1])))
          

    #update relevant high altitude indices with classification code
    if len(high_altitude_ind) > 0:
        #get relevant classification flag
        classification_flag_code = standard_classification_flag_name_to_classification_code['High Altitude - Meybeck Global Landform Classification']
        for inv in high_altitude_inds:
            data['classification'][inv].append(classification_flag_code)

    return data
    
###------------------------------------------------------------------------------------###

def near_coast_classification_by_GSFC(metadata,data):

    '''Check for stations near coast using the NASA Goddard Space Flight Center (GSFC) Ocean Color Group proximity to coast gridded dataset
       Negative distances represent locations over land (including land-locked bodies of water), while positive distances represent locations over the ocean. 
       There is an uncertainty of up to 1 km in the computed distance at any given point.
       if the absolute proximity to the coast is <50 km, then a station is classed as near coast.
    '''

    #load metadata
    GSFC_coastal_proximity = metadata['GSFC_coastline_proximity']
    update_metadata_stamp = metadata['meta_update_stamp']

    #define list for keeping indices associated with near coast stations
    near_coast_inds = np.empty(0, dtype=np.uint32)  
    
    #iterate through all GSFC coastal proximities
    for ii in range(len(GSFC_coastal_proximity)):

        #is absolute coastal proximity < 50km?
        if np.abs(GSFC_coastal_proximity[ii]) < 50:
            #is it last index? 
            if ii == (len(GSFC_coastal_proximity)-1):
                near_coast_inds = np.append(near_coast_inds,np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), len(data['time'])))
            else:
                near_coast_inds = np.append(near_coast_inds,np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), bisect.bisect_left(data['time'], update_metadata_stamp[ii+1])))
     
    #update relevant near coast indices with classification code 
    if len(near_coast_inds) > 0:
        #get relevant classification flag code
        classification_flag_code = standard_classification_flag_name_to_classification_code['Near Coast - GSFC']
        for inv in near_coast_inds:
            data['classification'][inv].append(classification_flag_code)

    return data

###------------------------------------------------------------------------------------###

def rural_urban_classification_by_standardised_metadata(metadata,data):
    
    '''Function that splits station into 'urban', 'rural' and 'unclassified' categories, derived from standardised network provided metadata, by 'lenient' and 'strict' definitions. 

       Fundamentally a line needs to be drawn to separate stations which make measurements which are representative in a 'rural' or 'urban' sense.
       This line is clearly subjective, but the line that is attempted to be drawn is that all measurements that are 'rural' are representative across measurement scales of >= 10km, and are located away from areas of significant human influence. 
       Stations which are flagged as being dominated by point source emissions are immediately considered 'urban'.
       Measurements made in the very largest of cities should clearly be classified as 'urban', but can be representative across large spatial scales (i.e. > 10km), thus the classification made should attempt to account for both the spatial representativity of measurements and the physical location of a station. 
 
       Two separate 'lenient' and 'strict' classifications are made to differentiate the characterisation what is a 'rural' or 'urban' station based on different classification rules. 
       The "strict" case requires more metadata to make a classification, the trade off in this case is that more stations end up being unclassified.

       The exact procedure for the classifications are as follows:

       ---------------------------------------------

       Lenient check definitions:

       -- If standardised_network_provided_station_classification == 'point_source' (and any subdivisions), station is immediately 'urban', and no further metadata is taken into account in classification   

       -- Attempt to next classify by both spatial representativeness of measurements and network provided station area classification
          - Spatial representativeness of measurements is defined by priority using the representative_radius, if this is not available it is defined by standardised_network_provided_measurement_scale
          - Station area classification is defined by standardised_network_provided_area_classification

       -- The determined classes for spatial representativeness and station area classification are combined to give a final classification.
         - If a station has conflicting spatial representativeness and area classification classes (e.g. 'rural' and 'urban'), the station is unclassified. 

       The classification groupings are as follows:
       --------
       REPRESENTATIVITY OF MEASUREMENTS:
       PRIORITY:
       representative_radius                                  urban:  < 10km                                                                            urban:  >= 10km

       SECONDARY:
       standardised_network_provided_measurement_scale:       urban: ['micro', 'middle', 'neighbourhood', 'city']                                       rural: ['regional']

       --------
       STATION CLASSIFICATIONS:
       standardised_network_provided_area_classification:     urban: ['urban', 'urban-centre', 'urban-suburban', 'rural-near_city']                     rural: ['rural', 'rural-regional', 'rural-remote']

       ---------------------------------------------

       Stict check definitions:

       -- The 'strict' check differs from the 'lenient' check by requiring that unless the station is flagged as being dominated by point-source emissions, both station area classification and spatial representativeness metadata is required to make a classification, otherwise a station will be unclassified.
    '''

    #load metadata
    area_classifications = metadata['standardised_network_provided_area_classification']
    station_classifications = metadata['standardised_network_provided_station_classification']
    representative_radii = metadata['representative_radius']
    measurement_scales = metadata['standardised_network_provided_measurement_scale']
    update_metadata_stamp = metadata['meta_update_stamp']
    
    #define lists for storing urban/rural defined indices 
    rural_lenient_inds = np.empty(0, dtype=np.uint32)
    urban_lenient_inds = np.empty(0, dtype=np.uint32)
    unclassified_lenient_inds = np.empty(0, dtype=np.uint32)
    rural_strict_inds = np.empty(0, dtype=np.uint32)
    urban_strict_inds = np.empty(0, dtype=np.uint32)
    unclassified_strict_inds = np.empty(0, dtype=np.uint32)
    
    #iterate through standard classifications
    for ii in range(len(area_classifications)):

        #get all data indices in metadata relevant period
        if ii == (len(area_classifications)-1):
            relevant_data_indices = np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), len(data['time']))
        else:
            relevant_data_indices = np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), bisect.bisect_left(data['time'], update_metadata_stamp[ii+1]))

        #Is the station flagged as being dominantly influenced by 'point-source' emissions? (using standardised_network_provided_station_classification field)
        if 'point_source' in str(station_classifications[ii]):
        #if so, classify as 'urban' (by lenient and strict definitions)
            urban_lenient_inds = np.append(urban_lenient_inds, relevant_data_indices)
            urban_strict_inds = np.append(urban_strict_inds, relevant_data_indices)
            continue
            
        #if not flagged as being dominated by point source emissions, now account for spatial representativity of measurements and area classification
        else:
            #create a dictionary for storing 'spatial_representativity_value', 'spatial_representativity_class', 'area_classification' (by default empty strings)
            class_dict = {'spatial_representativity_value':'', 'spatial_representativity_class':'', 'area_classification':''}
                
            #first attempt to define 'spatial_representativity_value' using 'representative radius' field
            if pd.isnull(representative_radii[ii]) == False:
                class_dict['spatial_representativity_value'] = representative_radii[ii]
            #if field not available, next attempt to define 'spatial_representativity_class' by "standardised_network_provided_measurement_scale" field
            if pd.isnull(measurement_scales[ii]) == False:
                class_dict['spatial_representativity_class'] = measurement_scales[ii]

            #now attempt to define 'area_classification' using 'standardised_network_provided_area_classification' field 
            if pd.isnull(area_classifications[ii]) == False:
                class_dict['area_classification'] = area_classifications[ii]

            #now dictionary is complete, make classifications based on 'lenient' and 'strict' rules
            #---------------------------------------------------------------------#
        
            #variables for setting if station is 'rural', 'urban' or 'unclassified': defined by spatial representativity, and area classification 
            defined_class_represent = ''
            defined_class_area = ''
            defined_class = ''

            #first attempt define by spatial representativity
            if class_dict['spatial_representativity_value'] != '':
                #urban
                if class_dict['spatial_representativity_value'] < 10.0:    
                    defined_class_represent = 'urban'
                #rural        
                elif class_dict['spatial_representativity_value'] >= 10.0:
                    defined_class_represent = 'rural'
              
            elif class_dict['spatial_representativity_class'] != '':
                #urban
                if class_dict['spatial_representativity_class'] in [u'micro', u'middle', u'neighbourhood', u'city']:    
                    defined_class_represent = 'urban'
                #rural        
                if class_dict['spatial_representativity_class'] in [u'regional']:
                    defined_class_represent = 'rural'    
            else:
                #unclassified
                    defined_class_represent = 'unclassified'    

            #next attempt to define by area classification 
            if class_dict['area_classification'] != '':
                #urban
                if class_dict['area_classification'] in ['urban', 'urban-centre', 'urban-suburban', 'rural-near_city']:
                    defined_class_area = 'urban' 
                #rural  
                elif class_dict['area_classification'] in ['rural', 'rural-regional', 'rural-remote']:
                    defined_class_area = 'rural' 
            else:
                #unclassified
                defined_class_area = 'unclassified'  

            #combine separately determined classifications to set final classification, by both 'lenient' and 'strict' rules
            #-----------------------------------------#
            #LENIENT CLASSIFICATION
            #-----------------------------------------#
        
            #urban
            if (defined_class_represent == 'urban') and (defined_class_area == 'unclassified'):
                defined_class = 'urban'
            elif (defined_class_represent == 'unclassified') and (defined_class_area == 'urban'):
                defined_class = 'urban'
            elif (defined_class_represent == 'urban') and (defined_class_area == 'urban'):
                defined_class = 'urban'
            #rural
            elif (defined_class_represent == 'rural') and (defined_class_area == 'unclassified'):
                defined_class = 'rural'
            elif (defined_class_represent == 'unclassified') and (defined_class_area == 'rural'):
                defined_class = 'rural'
            elif (defined_class_represent == 'rural') and (defined_class_area == 'rural'):
                defined_class = 'rural'
            #unclassified
            else:
                defined_class = 'unclassified'       
  
            #append relevant data indices
            if defined_class == 'urban':
                urban_lenient_inds = np.append(urban_lenient_inds, relevant_data_indices)
            elif defined_class == 'rural':
                rural_lenient_inds = np.append(rural_lenient_inds, relevant_data_indices)
            elif defined_class == 'unclassified':
                unclassified_lenient_inds = np.append(unclassified_lenient_inds, relevant_data_indices)

            #-----------------------------------------#
            #STRICT CLASSIFICATION
            #-----------------------------------------# 

            #urban
            if (defined_class_represent == 'urban') and (defined_class_area == 'urban'):
                defined_class = 'urban'
            #rural
            elif (defined_class_represent == 'rural') and (defined_class_area == 'rural'):
                defined_class = 'rural'
            #unclassified
            else:
                defined_class = 'unclassified'       
  
            #append relevant data indices
            if defined_class == 'urban':
                urban_strict_inds = np.append(urban_strict_inds, relevant_data_indices)
            elif defined_class == 'rural':
                rural_strict_inds = np.append(rural_strict_inds, relevant_data_indices)
            elif defined_class == 'unclassified':
                unclassified_strict_inds = np.append(unclassified_strict_inds, relevant_data_indices)
        
    #---------------------------------------------------------------------#
 
    #update relevant invalid inds with classification codes 
    if len(rural_lenient_inds) > 0:
        #get relevant classification flag code
        classification_flag_code = standard_classification_flag_name_to_classification_code['Rural Station - Lenient Metadata Derived']
        for inv in rural_lenient_inds:
            data['classification'][inv].append(classification_flag_code)
            
    if len(urban_lenient_inds) > 0:
        #get relevant classification flag code
        classification_flag_code = standard_classification_flag_name_to_classification_code['Urban Station - Lenient Metadata Derived']
        for inv in urban_lenient_inds:
            data['classification'][inv].append(classification_flag_code)
            
    if len(unclassified_lenient_inds) > 0:
        #get relevant classification flag code
        classification_flag_code = standard_classification_flag_name_to_classification_code['Unclassified Station - Lenient Metadata Derived']
        for inv in unclassified_lenient_inds:
            data['classification'][inv].append(classification_flag_code)
            
    if len(rural_strict_inds) > 0:
        #get relevant classification flag code
        classification_flag_code = standard_classification_flag_name_to_classification_code['Rural Station - Strict Metadata Derived']
        for inv in rural_strict_inds:
            data['classification'][inv].append(classification_flag_code)
            
    if len(urban_strict_inds) > 0:
        #get relevant classification flag code
        classification_flag_code = standard_classification_flag_name_to_classification_code['Urban Station - Strict Metadata Derived']
        for inv in urban_strict_inds:
            data['classification'][inv].append(classification_flag_code)
            
    if len(unclassified_strict_inds) > 0:
        #get relevant classification flag code
        classification_flag_code = standard_classification_flag_name_to_classification_code['Unclassified Station - Strict Metadata Derived']
        for inv in unclassified_strict_inds:
            data['classification'][inv].append(classification_flag_code)
    
    return data
    
###------------------------------------------------------------------------------------###

def rural_urban_classification_by_anthrome(metadata,data,class_resolution):

    '''Function that splits station into 'urban' and 'rural' categories, derived from University of Maryland Baltimore County (UMBC) ANTHROME gridded classifications describing the anthropogenic land use.
       All potential UMBC Anthrome classifications are split between 'Urban' and 'Rural'. 
       Depending on the input to the function, the Anthrome classifications used are derived from native 0.0833 x 0.0833 degree resolution, modal classification in a 5km perimeter around the station location, or modal classification in a 25km perimeter around the station location  

       The anthrome classifications are made as follows:
       1. The anthrome classifications are first stratified between six population density classes of differing orders of magnitude:
       -- urban (> 2500 persons km-2)
       -- dense (> 100 persons km-2)
       -- residential (10–100 persons km-2)
       -- populated (1–10 persons km-2)
       -- remote (< 1 person km-2)
       -- wild (0 persons km-2)
       2. Then anthromes are then classified by applying a sequence of classification thresholds to global gridded data for land area covered by urban settlements, rice, irrigated and rainfed crops and pastures.

       It is decided after empirical testing and evaluation to define all dense settlements and villages as 'urban', and the rest as 'rural'  
    '''

    #define 'Rural' UMBC classes
    rural_anthrome_classes = [u'residential irrigated croplands', u'residential rainfed croplands', u'populated croplands', u'remote croplands', u'residential rangelands', u'populated rangelands', u'remote rangelands', u'residential woodlands', u'populated woodlands', u'remote woodlands', u'inhabited treeless and barren lands', u'wild woodlands', u'wild treeless and barren lands', u'ocean']

    #define 'Urban' UMBC classes
    urban_anthrome_classes = [u'urban', u'mixed settlements', u'rice villages', u'irrigated villages', u'rainfed villages', u'pastoral villages']

    #load UMBC anthrome classes - the ones loaded are decided by input to the function
    #also set the classification rural/urban definition flags 
    #load native resolution derived classifications
    if class_resolution == 'native':
        umbc_anthrome_classifications = metadata['UMBC_anthrome_classification']
        rural_flag = 'Rural Station - Anthrome (Native Resolution)'
        urban_flag = 'Urban Station - Anthrome (Native Resolution)'
    #load modal 5km classifications
    elif class_resolution == '5km':
        umbc_anthrome_classifications = metadata['UMBC_modal_anthrome_classification_5km']
        rural_flag = 'Rural Station - Anthrome (Mode in 5km Perimeter)'
        urban_flag = 'Urban Station - Anthrome (Mode in 5km Perimeter)'
    #load modal 25km classifications
    elif class_resolution == '25km':
        umbc_anthrome_classifications = metadata['UMBC_modal_anthrome_classification_25km']   
        rural_flag = 'Rural Station - Anthrome (Mode in 25km Perimeter)'
        urban_flag = 'Urban Station - Anthrome (Mode in 25km Perimeter)'

    #load temporal updata metadata
    update_metadata_stamp = metadata['meta_update_stamp']

    #define lists for storing urban/rural defined indices
    urban_anthrome_inds = np.empty(0, dtype=np.uint32)
    rural_anthrome_inds = np.empty(0, dtype=np.uint32)
    
    #iterate through UMBC anthrome classes
    for ii in range(len(umbc_anthrome_classifications)):
        
        #is classification in defined 'rural' list'?
        if umbc_anthrome_classifications[ii] in rural_anthrome_classes:
            #is it last index? 
            if ii == (len(umbc_anthrome_classifications)-1):
                rural_anthrome_inds = np.append(rural_anthrome_inds,np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), len(data['time'])))
            else:
                rural_anthrome_inds = np.append(rural_anthrome_inds,np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), bisect.bisect_left(data['time'], update_metadata_stamp[ii+1])))
          
        #is classification in defined 'urban' list'?
        if umbc_anthrome_classifications[ii] in urban_anthrome_classes:
            #is it last index? 
            if ii == (len(umbc_anthrome_classifications)-1):
                urban_anthrome_inds = np.append(urban_anthrome_inds,np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), len(data['time'])))
            else:
                urban_anthrome_inds = np.append(urban_anthrome_inds,np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), bisect.bisect_left(data['time'], update_metadata_stamp[ii+1])))
          
    #update relevant rural defined inds with classification codes 
    if len(rural_anthrome_inds) > 0:
        #get relevant classification flag code
        classification_flag_code = standard_classification_flag_name_to_classification_code[rural_flag]
        for inv in rural_anthrome_inds:
            data['classification'][inv].append(classification_flag_code)

    #update relevant urban defined inds with classification codes 
    if len(urban_anthrome_inds) > 0:
        #get relevant classification flag code
        classification_flag_code = standard_classification_flag_name_to_classification_code[urban_flag]
        for inv in urban_anthrome_inds:
            data['classification'][inv].append(classification_flag_code)

    return data

###------------------------------------------------------------------------------------###

def rural_urban_classification_by_TOAR(metadata,data):

    '''Function that splits station into 'urban', 'rural' and 'unclassified' categories, following definitions by the Tropospheric Ozone Assessment Report (TOAR).'''

    #load metadata
    nighttime_lights = metadata['NOAA-DMSP-OLS_v4_nighttime_stable_lights']
    update_metadata_stamp = metadata['meta_update_stamp']

    #define lists for storing urban/rural/unclassified defined indices
    rural_TOAR_inds = np.empty(0, dtype=np.uint32)
    urban_TOAR_inds = np.empty(0, dtype=np.uint32)
    unclassifed_TOAR_inds = np.empty(0, dtype=np.uint32)

    #iterate through metadata
    #for ii in range(len(nighttime_lights)):
        
        #if nighttime_lights[ii] == '':
            #is it last index? 
        #    if ii == (len(nighttime_lights)-1):
        #        invalid_inds = np.append(invalid_inds,np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), len(data['time'])))
        #    else:
        #        invalid_inds = np.append(invalid_inds,np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), bisect.bisect_left(data['time'], update_metadata_stamp[ii+1])))
          
    #update relevant rural inds with classification codes 
    if len(rural_TOAR_inds) > 0:
        #get relevant classification flag code
        classification_flag_code = standard_classification_flag_name_to_classification_code['Rural Station - TOAR']
        for inv in rural_TOAR_inds:
            data['classification'][inv].append(classification_flag_code)

    #update relevant urban inds with classification codes  
    if len(urban_TOAR_inds) > 0:
        #get relevant classification flag code
        classification_flag_code = standard_classification_flag_name_to_classification_code['Urban Station - TOAR']
        for inv in urban_TOAR_inds:
            data['classification'][inv].append(classification_flag_code)
            
    #update relevant unclassified inds with classification codes 
    if len(unclassified_TOAR_inds) > 0:
        #get relevant classification flag code
        classification_flag_code = standard_classification_flag_name_to_classification_code['Unclassified Station - TOAR']
        for inv in unclassified_TOAR_inds:
            data['classification'][inv].append(classification_flag_code)

    return data

###------------------------------------------------------------------------------------###

def rural_classification_by_Joly_Peuch(metadata,data):

    '''function that splits station into 'rural' and 'unclassified' categories, following definitions by Joly and Peuch (2012)'''

    #load metadata
    update_metadata_stamp = metadata['meta_update_stamp']

    #define lists for storing rural/unclassified defined indices
    rural_Joly_Peuch_inds = np.empty(0, dtype=np.uint32)
    unclassifed_Joly_Peuch_inds = np.empty(0, dtype=np.uint32)

    #iterate through metadata
    #for ii in range(len()):
        
        #if [ii] == '':
            #is it last index? 
        #    if ii == (len()-1):
        #        invalid_inds = np.append(invalid_inds,np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), len(data['time'])))
        #    else:
        #        invalid_inds = np.append(invalid_inds,np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[ii]), bisect.bisect_left(data['time'], update_metadata_stamp[ii+1])))
          

    #update relevant rural inds with classification codes 
    if len(rural_Joly_Peuch_inds) > 0:
        #get relevant classification flag code
        classification_flag_code = standard_classification_flag_name_to_classification_code['Rural Station - Joly-Peuch']
        for inv in rural_Joly_Peuch_inds:
            data['classification'][inv].append(classification_flag_code)
            
    #update relevant unclassified inds with classification codes 
    if len(unclassified_Joly_Peuch_inds) > 0:
        #get relevant classification flag code
        classification_flag_code = standard_classification_flag_name_to_classification_code['Unclassified Station - Joly-Peuch']
        for inv in unclassified_Joly_Peuch_inds:
            data['classification'][inv].append(classification_flag_code)

    return data

###------------------------------------------------------------------------------------###

def get_standardised_metadata_classfication(field_name,classification_flag_name_root,metadata,data):

    '''Set appropriate classification flags for all values of a specific standardised metadata classification field.'''

    #load metadata
    field_metadata = metadata[field_name]
    update_metadata_stamp = metadata['meta_update_stamp']
    
    #iterate through all metadata values for field
    for metadata_ii, metadata_value in enumerate(field_metadata):

        #is metadata value valid?
        if pd.isnull(metadata_value) == False:

            #make exception for land use where all subgroups are dropped
            if (field_name == 'standardised_network_provided_land_use') or (field_name == 'MODIS_MCD12C1_v6_IGBP_land_use'):
                metadata_value = metadata_value.split('-')[0]

            #get relevant classification flag code
            classification_flag_code = standard_classification_flag_name_to_classification_code["{} = '{}'".format(classification_flag_name_root, metadata_value)]

            #get relevant indexes to write classification flags
            #is it last index? 
            if metadata_ii == (len(field_metadata)-1):
                class_inds = np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[metadata_ii]), len(data['time']))
            else:
                class_inds = np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[metadata_ii]), bisect.bisect_left(data['time'], update_metadata_stamp[metadata_ii+1]))
     
            #write classification codes
            for inv in class_inds:
                data['classification'][inv].append(classification_flag_code)
                  
    return data

###------------------------------------------------------------------------------------###

def spatial_representativity_classfication(metadata,data):

    '''Set spatial representativity classification flags from all pairs of representative radius/measurement scale metadata.
       This is done by first trying to classify spatial representativity using representative radius radius metadata.
       If this is not provided, then it is attempted to be classified using measurement scale metadata.
    '''

    #load metadata
    representative_radius_metadata = metadata['representative_radius']
    measurement_scale_metadata = metadata['standardised_network_provided_measurement_scale']
    update_metadata_stamp = metadata['meta_update_stamp']

    #iterate through metadata pairs
    for metadata_ii, representative_radius in enumerate(representative_radius_metadata):
        measurement_scale = measurement_scale_metadata[metadata_ii]

        #initialise classification_flag_code as Nan
        classification_flag_code = np.NaN

        #have valid representative radius?
        if pd.isnull(representative_radius) == False:
            #if so, get relevant classification flag code
            #<0.1km    
            if representative_radius < 0.1:
                classification_flag_code = standard_classification_flag_name_to_classification_code["Spatial Representativity Metadata = '<0.1km'"]
            #>=0.1km & <0.5 km
            elif (representative_radius >= 0.1) & (representative_radius < 0.5):
                classification_flag_code = standard_classification_flag_name_to_classification_code["Spatial Representativity Metadata = '0.1-0.5km'"]
            #>=0.5km & <4.0 km
            elif (representative_radius >= 0.5) & (representative_radius < 4.0):
                classification_flag_code = standard_classification_flag_name_to_classification_code["Spatial Representativity Metadata = '0.5-4km'"]
            #>=4.0km & <50.0 km
            elif (representative_radius >= 4.0) & (representative_radius < 50.0):
                classification_flag_code = standard_classification_flag_name_to_classification_code["Spatial Representativity Metadata = '4-50km'"]
            #>=50.0 km
            elif representative_radius >= 50.0:
                classification_flag_code = standard_classification_flag_name_to_classification_code["Spatial Representativity Metadata = '>=50km'"]

        #if not, have valid measurement scale?
        elif pd.isnull(measurement_scale) == False:
            #if so, get relevant classification flag code
            #<0.1km ~= 'micro'   
            if measurement_scale == 'micro':
                classification_flag_code = standard_classification_flag_name_to_classification_code["Spatial Representativity Metadata = '<0.1km'"]
            #>=0.1km & <0.5 km ~= 'middle'  
            elif measurement_scale == 'middle':
                classification_flag_code = standard_classification_flag_name_to_classification_code["Spatial Representativity Metadata = '0.1-0.5km'"]
            #>=0.5km & <4.0 km ~= 'neighbourhood'  
            elif measurement_scale == 'neighbourhood':
                classification_flag_code = standard_classification_flag_name_to_classification_code["Spatial Representativity Metadata = '0.5-4km'"]
            #>=4.0km & <50.0 km ~= 'city'  
            elif measurement_scale == 'city':
                classification_flag_code = standard_classification_flag_name_to_classification_code["Spatial Representativity Metadata = '4-50km'"]
            #>=50.0 km ~= 'regional'  
            elif measurement_scale == 'regional':
                classification_flag_code = standard_classification_flag_name_to_classification_code["Spatial Representativity Metadata = '>=50km'"]

        #have a valid classification data flag
        if pd.isnull(classification_flag_code) == False:
            #get relevant indexes to write classification flags
            #is it last index? 
            if metadata_ii == (len(representative_radius_metadata)-1):
                class_inds = np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[metadata_ii]), len(data['time']))
            else:
                class_inds = np.arange(bisect.bisect_left(data['time'], update_metadata_stamp[metadata_ii]), bisect.bisect_left(data['time'], update_metadata_stamp[metadata_ii+1]))
     
            #write classification codes
            for inv in class_inds:
                data['classification'][inv].append(classification_flag_code)

    return data